#!/usr/bin/python

from __future__ import absolute_import

from lxml import etree
from operator import attrgetter
from nidaba import storage


class Rect(object):
    """
    Native python replacement for gameras C++ Rect object.
    """

    def __init__(self, ul=(0, 0), lr=(0, 0)):
        self.ul = ul
        self.lr = lr
        self.lr_x = lr[0]
        self.lr_y = lr[1]
        self.ll_x = ul[0]
        self.ll_y = lr[1]
        self.ul_x = ul[0]
        self.ul_y = ul[1]
        self.ur_x = lr[0]
        self.ur_y = ul[1]


class hocrWord(object):
    """
    Dummy class associating word text with a bbox.
    """


class hocrLine(object):
    """
    Dummy class associating lines, words with their text and bboxes.
    """


def parse_bbox(prop_str):
    """
    Parses the microformat property string in the hOCR title field.

    Args:
        prop_str (unicode): The verbatim property string

    Returns:
        Rect: A rectangular area described in a bbox field

    Raises:
        ValueError: The property string either did not contain a bbox
                    definition or this definition was malformed.
    """
    for prop in prop_str.split(';'):
        p = prop.split()
        if p[0] == 'bbox':
            return Rect((p[1], p[2]), (p[3], p[4]))
        else:
            continue
    raise ValueError('bounding box not in proper format')


def get_hocr_lines_for_tree(treeIn):
    hocr_line_elements = treeIn.xpath("//html:span[@class='ocr_line'] |\
                                      //span[@class='ocr_line']",
                                      namespaces={'html':
                                                  "http://www.w3.org/1999/xhtml"})
    line_counter = 0
    lines_out = []
    all_words = []
    for hocr_line_element in hocr_line_elements:
        line_counter += 1
        words = hocr_line_element.xpath(".//html:span[@class='ocr_word'] |\
                                        .//span[@class='ocr_word'] ",
                                        namespaces={'html':
                                                    "http://www.w3.org/1999/xhtml"})
        words_out = []
        for word in words:
            aWord = hocrWord()
            aWord.text = ""
            if word.text:
                aWord.text += word.text
            # get rid of any inner elements, and just keep their text values
            for element in word.iterchildren():
                if element.text:
                    aWord.text += element.text
                word.remove(element)
            # set the contents of the xml element to the stripped text
            word.text = aWord.text
            aWord.bbox = parse_bbox(word.get('title'))
            aWord.element = word
            words_out.append(aWord)
        aLine = hocrLine()
        all_words = all_words + words_out
        aLine.words = words_out
        aLine.element = hocr_line_element
        aLine.bbox = parse_bbox(hocr_line_element.get('title'))
        lines_out.append(aLine)
    return lines_out, all_words


def close_enough(bbox1, bbox2, fudge=0.1):
    """
    Roughly matches two bboxes roughly using a fudge factor.

    Args:
        bbox1 (Rect): Rect object of a bounding box.
        bbox2 (Rect): Rect object of a bounding box.
        fudge (float): Fudge factor to account for slight variations in word
                       boundary detection between segmentation engines.

    Returns:
        bool: True if the bounding boxes are sufficiently aligned, False
        otherwise.
    """
    total_circum1 = (bbox1.lr_x - bbox1.ul_x) * 2 + \
        (bbox1.lr_y - bbox1.ul_y) * 2
    total_circum2 = (bbox1.lr_x - bbox1.ul_x) * 2 + \
        (bbox1.lr_y - bbox1.ul_y) * 2
    f = (total_circum1 + total_circum2) * fudge
    total_diff = (abs(bbox1.lr_x - bbox2.lr_x) + abs(bbox1.lr_y - bbox2.lr_y) +
                  abs(bbox1.ul_x - bbox2.ul_x) + abs(bbox1.ul_y - bbox2.ul_y))
    if total_diff < f:
        return True
    else:
        return False


def sort_words_bbox(words):
    """
    Sorts word bboxes of a document in European reading order (upper left to
    lower right). The list is sorted in place.

    Args:
        words (list): List of hocrWord object containing Rects in the field
                      bbox and the recognized text in the text field.

    Returns:
        list: The sorted word list.
    """
    words.sort(key=attrgetter('bbox.lr_y'))
    words.sort(key=attrgetter('bbox.lr_x'))
    words.sort(key=attrgetter('text'))
    return words


def score_word(lang, word):
    """
    A simple token scoring function similar to the one used in Bruce Robertsons
    rigaudon. FIXME: Actually score input tokens.

    Args:
        lang (unicode): Language to use for scoring.
        word (unicode): Input token to score

    Returns:
        int: Value representing the input tokens score. Higher values are
             closer to native language words.
    """
    # IN_DICT_SCORE = 1000
    # IN_DICT_LOWER_SCORE = 100
    # CAMEL_CASE_SCORE = 1
    # ALL_CAPS_SCORE = 10
    score_total = 0
    # no language => no score
    if not lang:
        return score_total
    # if spell(lang, word):
    #     score_total = score_total + IN_DICT_SCORE
    # elif spell(lang, word.lower()):
    #     score_total = score_total + IN_DICT_LOWER_SCORE
    # if score_total > 0:
    #     if word.istitle():
    #         score_total = score_total + CAMEL_CASE_SCORE
    #     elif word.isupper():
    #         score_total = score_total + ALL_CAPS_SCORE
    return score_total


def merge(docs, lang, output):
    """
    Merges multiple hOCR documents into a single one.

    First bboxes from all documents are roughly matched, then all matching
    bboxes are scored using a spell checker. If no spell checker is available
    all matches will be merged without ranking.

    The matching is naive, i.e. we just grab the first input document and
    assume that all other documents have similar segmentation results. Issues
    like high variance in segmentation, especially word boundaries are not
    accounted for.

    Args:
        docs (iterable): A list of storage tuples of input documents
        lang (unicode): A language identifier for the spell checker
        output (tuple): Storage tuple for the result

    Returns:
        tuple: The output storage tuple. Should be the same as ```output```.
    """
    parser = etree.HTMLParser()
    tree1 = etree.parse(storage.get_abs_path(docs[0][0], docs[0][1]), parser)
    lines_1, words_1 = get_hocr_lines_for_tree(tree1)
    sort_words_bbox(words_1)
    other_words = []
    for doc in docs[1:]:
        try:
            tree2 = etree.parse(storage.get_abs_path(doc[0], doc[1]), parser)
            lines_2, words_2 = get_hocr_lines_for_tree(tree2)
            other_words = other_words + words_2
        except Exception as e:
            print e

    sort_words_bbox(other_words)
    positional_lists = []
    positional_list = []
    x = 0

    # Make a list of positional_lists, that is alternatives for a given
    # position, skipping duplicate position-words
    while x < len(other_words):
        try:
            if len(positional_list) == 0:
                positional_list.append(other_words[x])
            else:
                if close_enough(other_words[x - 1].bbox, other_words[x].bbox):
                    # skip if the text is the same, so that we just get unique
                    # texts for this position
                    if not other_words[x - 1].text == other_words[x].text:
                        positional_list.append(other_words[x])
                else:
                    if not x == 0:
                        positional_lists.append(positional_list)
                        positional_list = []
        except IndexError:
            pass
        x = x + 1

    # we now have a list of list of unique words for each position
    # let's select from each the first one that passes spellcheck
    replacement_words = []

    # make a 'replacement_words' list with all of the best, non-zero-scoring
    # suggestions for each place
    for positional_list in positional_lists:
        for word in positional_list:
            word.score = score_word(lang, word.text)
        positional_list.sort(key=attrgetter('score'), reverse=True)
        if positional_list[0].score > 0:
            replacement_words.append(positional_list[0])

    # now replace the originals
    for word in words_1:
        for replacement_word in replacement_words:
            word.score = score_word(lang, word.text)
            if close_enough(word.bbox, replacement_word.bbox) and (
                    word.score < replacement_word.score):
                word.element.text = replacement_word.text

        for positional_list in positional_lists:
            print "##"
            for word in positional_list:
                print word.bbox, word.text

    storage.write_text(*output, text=etree.tostring(tree1.getroot(),
                                                    encoding='unicode'))
    return output



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile

from lxml import etree
from nose.plugins.skip import SkipTest
from mock import patch, MagicMock

thisfile = os.path.abspath(os.path.dirname(__file__))
resources = os.path.abspath(os.path.join(thisfile, 'resources/ocropus'))


class KrakenTests(unittest.TestCase):

    """
    Tests for the kraken plugin.
    """

    def setUp(self):
        try:
            config_mock = MagicMock()
            storage_path = unicode(tempfile.mkdtemp())
            config_mock.nidaba_cfg = {
                'storage_path': storage_path,
                'kraken_models': {'default': ('test', 'en-default.hdf5')},
                'ocropus_models': {'ocropus': ('test', 'en-default.pyrnn.gz')},
                'plugins_load': {}
            }
    
            self.patches = {
                'nidaba.config': config_mock,
            }
            self.patcher = patch.dict('sys.modules', self.patches)
            self.addCleanup(self.patcher.stop)
            self.patcher.start()
            self.storage_path = storage_path
            shutil.copytree(resources, self.storage_path + '/test')
            from nidaba.plugins import kraken
            kraken.setup()
            self.kraken = kraken
        except:
            raise SkipTest

    def tearDown(self):
        self.patcher.stop()
        shutil.rmtree(self.storage_path)


    def test_segmentation(self):
        """
        Test that kraken's page segmentation is working
        """
        o = self.kraken.segmentation_kraken.run(('test', 'image_png.png'))
        self.assertTrue(os.path.isfile(os.path.join(self.storage_path, *o[0])),
                        msg='Kraken did not output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *o[0])))
        except etree.XMLSyntaxError as e:
            print(e.message)
            self.fail(msg='The outpath was not valid xml!')


    def test_hdf5_model(self):
        """
        Test that kraken creates hocr output with HDF5 models.
        """
        ocr = self.kraken.ocr_kraken.run((('test', 'segmentation.xml'), 
                                      ('test', 'image_png.png')), 
                                     model='default')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError as e:
            print(e.message)
            self.fail(msg='The outpath was not valid html/xml!')


    def test_file_outpath_png(self):
        """
        Test that kraken creates hocr output for pngs.
        """
        ocr = self.kraken.ocr_kraken.run((('test', 'segmentation.xml'), 
                                      ('test', 'image_png.png')), 
                                     model='ocropus')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError as e:
            print(e.message)
            self.fail(msg='The outpath was not valid xml!')

    def test_file_outpath_tiff(self):
        """
        Test that kraken creates hocr output for tiffs.
        """
        ocr = self.kraken.ocr_kraken.run((('test', 'segmentation.xml'), 
                                      ('test', 'image_tiff.tiff')), 
                                     model='ocropus')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid xml!')


    def test_file_outpath_jpg(self):
        """
        Test that kraken creates hocr output for jpgs.
        """
        ocr = self.kraken.ocr_kraken.run((('test', 'segmentation.xml'), 
                                      ('test', 'image_jpg.jpg')), 
                                     model='ocropus')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The outpath was not valid xml!')


    def test_nlbin(self):
        """
        Test that kraken's nlbin is callable.
        """
        ret = self.kraken.nlbin.run(('test', 'image_jpg.jpg'))
        self.assertTrue(os.path.isfile(os.path.join(self.storage_path, *ret)),
                        msg='Kraken did not output a file!')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
nidaba.plugins
==============

"""

from __future__ import absolute_import, print_function

from nidaba.config import nidaba_cfg

import stevedore


def setup(ext, data):
    ext.plugin.setup(**data[ext.name])

mgr = stevedore.NamedExtensionManager(namespace='nidaba.plugins',
                                      names=nidaba_cfg['plugins_load'].keys(),
                                      propagate_map_exceptions=True)

# only initialize extensions if any are loaded
if mgr.extensions:
    mgr.map(setup, nidaba_cfg['plugins_load'])



# -*- coding: utf-8 -*-
"""
nidaba.tasks.util
~~~~~~~~~~~~~~~~~

Various utility tasks that either can't be classified by purpose or are one of
a kind.

"""

from __future__ import absolute_import

from nidaba.tasks.helper import NidabaTask
from nidaba.celery import app


@app.task(base=NidabaTask, name=u'nidaba.util.sync')
def sync(doc):
    """
    Takes ones argument and returns it. Used to synchronized stuff as
    chaining groups is not possible with the current celery version.

    Args:
        doc: An arbitrary input argument

    Returns:
        The input argument unaltered
    """
    return doc



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import unittest
import os
import shutil
import tempfile
import ctypes

from lxml import etree
from distutils import spawn
from nose.plugins.skip import SkipTest
from mock import patch, MagicMock

thisfile = os.path.abspath(os.path.dirname(__file__))
tessdata = os.path.abspath(os.path.join(thisfile, 'resources'))


class TesseractTests(unittest.TestCase):

    """
    Tests the tesseract plugin.
    """

    def setUp(self):
        self.config_mock = MagicMock()
        storage_path = unicode(tempfile.mkdtemp())
        self.config_mock.nidaba_cfg = {
            'storage_path': storage_path,
            'plugins_load': {}
        }

        self.patches = {
            'nidaba.config': self.config_mock,
        }
        self.patcher = patch.dict('sys.modules', self.patches)
        self.patcher2 = patch('nidaba.storage.nidaba_cfg', self.config_mock.nidaba_cfg)
        self.addCleanup(self.patcher2.stop)
        self.addCleanup(self.patcher.stop)
        self.patcher.start()
        self.patcher2.start()
	self.storage_path = storage_path
        shutil.copytree(tessdata, self.storage_path + '/test')
        from nidaba.plugins import tesseract
        self.tesseract = tesseract
        

    def tearDown(self):
        self.patcher.stop()
        shutil.rmtree(self.storage_path)

    def test_capi_multiple(self):
        """
        Test that tesseract CAPI calls create hocr output for multiple
        languages.
        """
        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest
        self.tesseract.setup(tessdata=tessdata, implementation='capi')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.tiff')),
                                           languages=['grc', 'eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            doc = etree.parse(open(os.path.join(self.storage_path, *ocr)))
            print(etree.tostring(doc))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_direct_multiple(self):
        """
        Test that direct tesseract calls create hocr output for multiple
        languages.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='direct')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.tiff')),
                                           languages=['grc', 'eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_capi_extended(self):
        """
        Test that the CAPI extended output contains character cuts in each
        ocr_line and character confidences in each ocrx_word.
        """

        try:
            ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest
        self.tesseract.setup(tessdata=tessdata, implementation='capi')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.tiff')),
                                           languages=['eng'],
                                           extended=True)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')

        try:
            h = etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')
        self.assertIsNotNone(h.findall(".//line"), msg='Tesseract did not write lines.')
        self.assertIsNotNone(h.findall(".//seg"), msg='Tesseract did not write segments.')
        self.assertIsNotNone(h.findall(".//g"), msg='Tesseract did not write graphemes.')


    def test_capi_file_output_png(self):
        """
        Test that tesseract CAPI calls create hocr output for pngs.
        """

        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='capi')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.png')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_capi_file_output_tiff(self):
        """
        Test that tesseract CAPI calls create hocr output for tiffs.
        """
        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='capi')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.tiff')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_capi_file_output_jpg(self):
        """
        Test that tesseract CAPI calls create hocr output for jpgs.
        """

        try:
            t = ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='capi')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.jpg')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_direct_file_output_png(self):
        """
        Test that direct tesseract calls create hocr output for pngs.
        """

        if not spawn.find_executable('tesseract'):
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='direct')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.png')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_direct_file_output_tiff(self):
        """
        Test that direct tesseract calls create hocr output for tiffs.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='direct')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.tiff')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


    def test_direct_file_output_jpg(self):
        """
        Test that direct tesseract calls create hocr output for jpgs.
        """
        if not spawn.find_executable('tesseract'):
            raise SkipTest

        self.tesseract.setup(tessdata=tessdata, implementation='direct')
        ocr = self.tesseract.ocr_tesseract.run((('test', 'segmentation.xml'), ('test',
                                           'image.jpg')),
                                           languages=['eng'],
                                           extended=False)
        outpath = os.path.join(self.storage_path, *ocr)
        self.assertTrue(os.path.isfile(outpath), msg='Tesseract did not '
                        'output a file!')
        try:
            etree.parse(open(os.path.join(self.storage_path, *ocr)))
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


if __name__ == '__main__':
    unittest.main()
